#include <Engine/types/scene/Scene.h>
#include <Engine/types/entity/Camera.h>
#include <Engine/types/entity/InstancedSprite.h>
#include <JGL/JGL.h>
#include <algorithm>

using namespace Engine;
bool Scene::EntityListContains(const Entity* entity) const {
    auto flat = GetFlatEntityList(entity_list);

    for (auto* e : flat)
        if (e == entity)
            return true;
    return false;
}

bool Scene::FixedListContains(const Fixed* fixed) const {
    for (auto* f : fixed_list)
        if (f == fixed)
            return true;
    return false;
}

size_t Scene::FixedCount() const {
    return fixed_list.size();
}

size_t Scene::EntityCount() const {
    return GetFlatEntityList(entity_list).size();
}

void Scene::Update() {
    if (active_camera)
        active_camera->Update();

    for (auto& e : entity_list)
        if (auto* c = dynamic_cast<Camera*>(e); c == nullptr)
            e->Update();
}

void Scene::Render(RenderTarget* render_target) {
    // The Camera and the HUD are ignored here because they're special cases.
    std::vector<Renderable*> display_list{};
    for (auto* e : GetFlatEntityList(entity_list))
        if (auto* r = dynamic_cast<Renderable*>(e))
            if (auto* c = dynamic_cast<Camera*>(r); c == nullptr)
                if (auto* h = dynamic_cast<Hud*>(c); h == nullptr)
                    display_list.push_back(r);

    std::sort(display_list.begin(), display_list.end(), [](Renderable* a, Renderable* b) {
        return a->GetLayer() > b->GetLayer();
    });

    J2D::Begin( render_target, true);

    if (active_camera)
        active_camera->Render();

    for (auto& f : fixed_list)
        if (f->Enabled())
            f->Render();

    for (auto& r : display_list)
        r->Render();

    if (heads_up_display)
        heads_up_display->Render();

    J2D::End();
}

Scene::~Scene() {
    for (auto* f : fixed_list)
        delete f;

    for (auto* e : entity_list)
        delete e;

    delete heads_up_display;
    delete active_camera;
}

void Scene::Unload() {
    for (auto* f : fixed_list)
        delete f;

    for (auto* e : entity_list)
        delete e;

    delete heads_up_display;
}

void Scene::AppendEntity(Entity* entity) {
    if (!EntityListContains(entity))
        entity_list.push_back(entity);
}

void Scene::AppendFixed(Fixed* fixed) {
    if (!FixedListContains(fixed))
        fixed_list.push_back(fixed);
}

void Scene::DestroyEntity(Entity *entity) {
    auto it = std::find(entity_list.begin(), entity_list.end(), entity);
    if (it != entity_list.end())
        delete *it, entity_list.erase(it);
}

void Scene::DestroyFixed(Fixed* fixed) {
    auto it = std::find(fixed_list.begin(), fixed_list.end(), fixed);
    if (it != fixed_list.end())
        delete *it, fixed_list.erase(it);
}

void Scene::RemoveEntity(Entity* entity) {
    auto it = std::find(entity_list.begin(), entity_list.end(), entity);
    if (it != entity_list.end())
        entity_list.erase(it);
}


void Scene::RemoveFixed(Fixed* fixed) {
    auto it = std::find(fixed_list.begin(), fixed_list.end(), fixed);
    if (it != fixed_list.end())
        fixed_list.erase(it);
}

std::string Scene::GetName() const {
    return name;
}

Camera* Scene::GetActiveCamera() const {
    return active_camera;
}

std::vector<Entity*> Scene::GetFlatEntityList(const std::vector<Entity*>& ent_list) const {
    std::vector<Entity*> result{};

    for (auto* e : ent_list) {
        auto children = GetFlatEntityList(e->GetChildren());
        for (auto* c : children)
            result.push_back(c);
        result.push_back(e);
    }
    return result;
}

Texture* Scene::GetInstancedTexture(const InstancedSprite* user) {
    for (auto* itx : instanced_textures)
        if (itx->InUseBy(user))
            return itx->GetTexture();

    auto* t = new Texture(user->GetTextureFilesystemPath());
    auto* itx = new InstancedTexture(t, user);
    instanced_textures.push_back(itx);
    return t;
}

Texture* Scene::GetInstancedAlphaMask(const InstancedSprite* user) {
    for (auto* itx : instanced_alpha_masks)
        if (itx->InUseBy(user))
            return itx->GetTexture();

    if (!user->GetAlphaMaskFilesystemPath().empty()) {
        auto *t = new Texture(user->GetTextureFilesystemPath());
        auto *itx = new InstancedTexture(t, user);
        instanced_textures.push_back(itx);
        return t;
    }
    return nullptr;
}

